import asyncio
import datetime
import logging
import json
import sys
from collections import OrderedDict


LOGGING_CONFIG_DEFAULTS = dict(
    version=1,
    disable_existing_loggers=False,

    filters={
        'no_keepalive_timeout': {
            '()': 'sanic_json_logging.formatters.NoKeepaliveFilter'
        }
    },

    root={
        'level': 'INFO',
        'handlers': ['console']
    },
    loggers={
        'root': {
            'level': 'INFO',
            'handlers': [],
            'propagate': True,
        },
        'sanic.error': {
            'level': 'INFO',
            'handlers': ['console'],
            'propagate': False,
            'qualname': 'sanic.error'
        },

        'sanic.access': {
            'level': 'INFO',
            'handlers': ['access_console'],
            'propagate': False,
            'qualname': 'sanic.access'
        }
    },
    handlers={
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'generic',
            'stream': sys.stdout,
            'filters': ['no_keepalive_timeout']
        },
        'access_console': {
            'class': 'logging.StreamHandler',
            'formatter': 'access',
            'stream': sys.stdout,
            'filters': ['no_keepalive_timeout']
        },
    },
    formatters={
        'generic': {
            'class': 'sanic_json_logging.formatters.JSONFormatter',
        },
        'access': {
            'class': 'sanic_json_logging.formatters.JSONReqFormatter'
        },
    }
)


# Gets rid of annoying info level messages where
# browser is sending connection keepalive header
class NoKeepaliveFilter(logging.Filter):
    def filter(self, record):
        try:
            return 'KeepAlive Timeout' not in record.msg
        except:  # noqa: E722
            return True


class JSONFormatter(logging.Formatter):
    def __init__(self, *args, context='context', **kwargs):
        super(JSONFormatter, self).__init__(*args, **kwargs)

        self._context_attr = context

    @staticmethod
    def format_timestamp(time):
        return datetime.datetime.utcfromtimestamp(time).isoformat() + 'Z'

    def format(self, record, serialize=True):
        try:
            msg = record.msg % record.args
        except TypeError:
            msg = record.msg

        try:
            msg_type = record.type
        except AttributeError:
            msg_type = 'log'

        # Deal with tracebacks
        exc = ''
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if exc[-1:] != '\n':
                exc += '\n'
            exc += record.exc_text
        if record.stack_info:
            if exc[-1:] != '\n':
                exc += '\n'
            exc += self.formatStack(record.stack_info)
        exc = exc.lstrip('\n').replace('\n', '<br>')

        message = OrderedDict((
            ('timestamp', self.format_timestamp(record.created)),
            ('level', record.levelname),
            ('message', msg),
            ('type', msg_type),
            ('logger', record.name),
        ))

        if msg_type == 'log':
            message['filename'] = record.filename
            message['lineno'] = record.lineno
        elif msg_type == 'event':
            message['event_type'] = record.event_type

        if len(exc) > 0:
            message['traceback'] = exc
            message['type'] = 'exception'
        if hasattr(record, 'data'):
            message['data'] = record.data

        try:
            current_task = asyncio.Task.current_task()
            # TODO make 'context' configurable by __init__(arg) from logging config
            if current_task and hasattr(current_task, 'context'):
                message['req_id'] = current_task.context.get('req_id', 'unknown')
        except Exception:
            pass

        # TODO log an error if json.dumps fails
        return json.dumps(message)


class JSONReqFormatter(JSONFormatter):
    def format(self, record, serialize=True):
        # Create message dict
        host = record.request.host
        try:
            if 'X-Forwarded-For' in record.request.headers:
                host = record.request.headers['X-Forwarded-For']
        except:  # noqa: E722
            pass

        message = OrderedDict((
            ('timestamp', self.format_timestamp(record.created)),
            ('level', record.levelname),
            ('method', record.request.method),
            ('path', record.request.path),
            ('remote', '{0}:{1}'.format(record.request.ip, record.request.port)),
            ('user_agent', record.request.headers['user-agent']),
            ('host', host),
            ('response_time', round(record.time, 2)),
            ('req_id', record.req_id),
            ('logger', record.name)
        ))

        if record.response is not None:  # not Websocket
            message['status_code'] = record.response.status
            if hasattr(record.response, 'body'):
                message['length'] = len(record.response.body)
            else:
                message['length'] = -1

            message['type'] = 'access'
        else:
            message['type'] = 'ws_access'

        # Can't remember why I added this
        if 'error_message' in record.request:
            try:
                message['request_info']['error_message'] = record.request['error_message']
            except KeyError:
                message['request_info'] = {'error_message': record.request['error_message']}

        return json.dumps(message)



import asyncio
import logging
import logging.config
import time
import uuid

from sanic_json_logging.formatters import LOGGING_CONFIG_DEFAULTS
from sanic_json_logging.sanic_app import NoAccessLogSanic


__version__ = '1.1.1'
__all__ = ['setup_json_logging', 'NoAccessLogSanic']


def setup_json_logging(app, configure_task_local_storage=True, context_var='context'):
    """
    Sets up request logging
    """
    # Set up logging
    LOGGING_CONFIG_DEFAULTS['formatters']['generic']['context'] = context_var
    logging.config.dictConfig(LOGGING_CONFIG_DEFAULTS)

    if configure_task_local_storage:
        # Set task factory
        asyncio.get_event_loop().set_task_factory(lambda loop, coro: _task_factory(loop, coro, context_var))

    req_logger = logging.getLogger('sanic.access')

    # Middleware to start a timer to gather request length.
    # Also generate a request ID, should really make request ID configurable
    @app.middleware('request')
    async def log_json_pre(request):
        """
        Setup unique request ID and start time
        :param request: Web request
        """
        current_task = asyncio.Task.current_task()
        if current_task:
            if hasattr(current_task, 'context'):
                current_task.context['req_id'] = str(uuid.uuid4())
                current_task.context['req_start'] = time.perf_counter()
            else:
                current_task.context = {
                    'req_id': str(uuid.uuid4()),
                    'req_start': time.perf_counter()
                }

    # This performs the role of access logs
    @app.middleware('response')
    async def log_json_post(request, response):
        """
        Calculate response time, then log access json
        :param request: Web request
        :param response: HTTP Response
        :return:
        """
        req_id = 'unknown'
        time_taken = -1

        current_task = asyncio.Task.current_task()
        if current_task and hasattr(current_task, 'context'):
            req_id = current_task.context['req_id']
            time_taken = time.perf_counter() - current_task.context['req_start']

        req_logger.info(None, extra={'request': request, 'response': response, 'time': time_taken, 'req_id': req_id})


def _task_factory(loop, coro, context_var='context') -> asyncio.Task:
    """
    Task factory function
    Fuction closely mirrors the logic inside of
    asyncio.BaseEventLoop.create_task. Then if there is a current
    task and the current task has a context then share that context
    with the new task
    """
    task = asyncio.Task(coro, loop=loop)
    if task._source_traceback:  # flake8: noqa
        del task._source_traceback[-1]  # flake8: noqa

    # Share context with new task if possible
    current_task = asyncio.Task.current_task(loop=loop)
    if current_task is not None and hasattr(current_task, context_var):
        setattr(task, context_var, current_task.context)

    return task



import sanic


class NoAccessLogSanic(sanic.Sanic):
    def __init__(self, *args, **kwargs):
        super(NoAccessLogSanic, self).__init__(*args, **kwargs)

        # The logo is cool... just not in my logs
        self.config.LOGO = None

    def run(self, *args, **kwargs):
        if 'access_log' not in kwargs:
            kwargs['access_log'] = False
        return super(NoAccessLogSanic, self).run(*args, **kwargs)

    def _helper(self, *args, **kwargs):
        result = super(NoAccessLogSanic, self)._helper(*args, **kwargs)
        result['access_log'] = False
        return result



#!/usr/bin/env python3
from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

setup(
    name='sanic-json-logging',
    version='1.1.1',
    description="Simple library to emit json formatted logs to stdout",
    long_description=readme,
    author="Terry Cain",
    author_email='terry@terrys-home.co.uk',
    url='https://github.com/terrycain/sanic-json-logging',
    packages=find_packages(include=['sanic_json_logging*']),
    include_package_data=True,
    install_requires=['sanic>=0.8.3'],
    license="Apache 2",
    zip_safe=False,
    keywords='sanic json logging',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    test_suite='tests'
)
